###################################################################
# FTS Implementation
###################################################################

The implementation provided here is described in the paper: "Efficient Static Analysis and Verification of Featured Transition Systems", submitted for publication in the special issue "Configurable Systems" of the "Empirical Software Engineering Journal" (December 2019).
This analysis program is written in python and makes use of pyhton bindings to the Z3 SAT solver. 
The python code is compatible with python versions 2.7.15 and 3.*
The implementation is structured in three parts:
 - the frontend, that allows to load an FTS written in a `.dot` format, on which to start the analysis.
 - the analysis, which implements the algorithms discussed in the paper and calls the Z3 solver
 - the Z3 solver calls that verify the presence of hidden deadlock states, false optional transitions, and dead transitions.

## Dependency

Our implementation depends on the `z3` python bindings for the analysis, and on the `pydot` and `lrparsing` python library for the frontend.
These packages can be installed with the following command line
```bash
$ pip install z3-solver pydot lrparsing
```

In case `pip` is not installed on your computer, follow [these instructions](#troubleshooting-installation-problems). 

## Structure

The implementation is structured in several files:
 - *gzoumUtils.py*: this file contains some utility definitions
 - *gzoumLogic.py*: this file contains a simple implementation of Boolean formula, with a method to check the satisfiability of these formula, using the Z3 solver
 - *gzoumGraph.py*: this file contains a generic implementation of graphs, upon which our FTS implementation is built
 - *gzoumFTS.py*: this file contains our implementation of FTSs, with the algorithm that compute the different criteria for nodes and transitions discussed in the paper
 - *analyser.py*: this file implements the frontend of our tool

This implementation is distributed under the GPLv3 license.


###################################################################
## Usage
###################################################################

As discussed before, the frontend of our tool loads an FTS encoded in a `.dot` format and analyses it.
The command line to use the frontend is as follows

```bash
$ python3 analyser.py --hdead [fts_to_analyse.dot]   ##  Liveness-only analysis
$ python3 analyser.py [--full] [fts_to_analyse.dot]  ##  Complete analysis
```

The FTS models in the `.dot` format are in the directory "FTSmodels".

Then previous implementation, presented as artifact at the SPLC 2019 conference,
can be found at:   https://doi.org/10.5281/zenodo.3229671


###################################################################
## Dot Representation of FTS Models
###################################################################

The .dot format is one of the graphical notations supported by the graphviz environment
(see  graphviz.org). 
Several online viewers exists for this format.
(e.g. https://dreampuf.github.io/GraphvizOnline)

We illustrate our encoding of an FTS with the vending machine example:
```dot
digraph G { 
  # the FTS is encoded as a directed graph
  # the following three lines are only used for display
  node [shape=circle];
  rankdir=LR;
  FeatureModel [shape=plaintext, style=filled, color=yellow, label="FM = s or t"];
  #
  FM="s or t";    # the feature model of the FTS
  name="vending"; # the name of the FTS
  #
  1 [initial=True] # states that the initial state of the FTS is "1"
  #
  # all the transitions of the FTS
  # the label of a transition is structured in two parts, separated with the "|" symbols:
  #  - the first part is the action of the transition
  #  - the second part is the guard of the transition
  #
  1 -> 2 [label="pay | not f"]
  1 -> 3 [label="free | f"]
  2 -> 3 [label="change | not f"]
  3 -> 4 [label="cancel | c"]
  3 -> 5 [label="soda | s"]
  3 -> 6 [label="tea | t"]
  4 -> 1 [label="return | c"]
  5 -> 7 [label="serveSoda | s"]
  6 -> 7 [label="serveTea | t"]
  7 -> 8 [label="open | not f"]
  7 -> 1 [label="take | f"]
  8 -> 9 [label="take | not f"]
  9 -> 1 [label="close | not f"]
}
```

The syntax of the guard (and of the feature model) is a simple Boolean expression, with standard Boolean operators (`not`, `and`, `xor`, `or`, `=>` and `<=>`), `True` and `False`, parenthesis and variables.


###################################################################
## Troubleshooting Installation Problems
###################################################################

In our experience, the installation of our analyser's dependencies can fail for two reasons:
 * `pip` is not installed
 * on windows, some version of `z3-solver` requires some visual studio components which are not installed on your computer

## `pip` is not installed

Depending on your system, there are different way to install pip.
Here is a list of what we tested.

## MacOS

In our experience, the simplest way to install pip on MacOS is to install switch to [python3](https://www.python.org/downloads/), which includes pip.
Then, the python (and respectively pip) tool takes the name `python3` (resp. `pip3`).
Hence the command line to install the dependencies becomes
```
pip3 install z3-solver pydot lrparsing
```
and using our analyser becomes
```bash
$ python3 analyser.py [fts_to_analyse.dot]
```

## Debian/Ubuntu Linux distribution

The tool `pip` is available in the package `python-pip`.

## Gentoo Linux distribution

The tool `pip` is available in the package `dev-python/pip`.
In gentoo, pip packages are installed locally for each user, and so the command line to install the dependencies becomes
```
pip install --user z3-solver pydot lrparsing
```

## Windows

Recent versions of `python2.7` come with a module that can install `pip`.
You need to open a console and execute
```
python -m ensurepip
```

Moreover, all versions of `python3.*` include the `pip` tool.

## Installing z3-solver on Windows

If `pip install z3-solver` fails, you can install it by hand with the followins steps:
1. download the version of `z3` from [here](https://github.com/Z3Prover/z3/releases) corresponding to your architecture
2. unzip it at any `${z3_location}`
3. go in your Windows settings (or control panel), and change some system variables:
    1. add to your `PATH` variable the path `${z3_location}\bin`
    2. add to your `PYTHONPATH` variable (create it if it does not exist yet) the path `${z3_location}\bin\python`
