#!/usr/bin/python

__author__ = "Michael Lienhardt"
__copyright__ = "Copyright 2019, Michael Lienhardt"
__license__ = "GPL3"
__version__ = "1"
__maintainer__ = "Michael Lienhardt"
__email__ = "michael lienhardt@laposte.net"
__status__ = "Prototype"



from gzoumUtils import container
import gzoumLogic as gzl


#
# CORE IMPLEMENTATION OF FTS
#

class Transition(object):
  """Generic edge class specification"""
  __slots__ = '_nodes', '_data', '_visited'
  def __init__(self, data, _in, _out):
    self._nodes = (_in, _out)
    self._data = data
    self._visited = False

  def get_data(self): return self._data
  def get_in(self): return self._nodes[0]
  def get_out(self): return self._nodes[1]


class State(object):
  """Simple implementation of a state class"""
  __slots__ = '_id', '_data', '_visited', '_connections'
  def __init__(self, id, data):
    self._id = id
    self._data = data
    self._visited = False
    self._connections = (set(), set())
  def prevs(self): return self._connections[0]
  def nexts(self): return self._connections[1]


class FTS(object):
  __slots__ = '_nodes', '_connections',  '_fm', '_initial', 'HIDDEN_DEADLOCKS', 'DEAD_TRANSITIONS'

  def __init__(self, fm):
    self._nodes = {}
    self._connections = []
    self._fm = fm                # the feature model of the FTS
    self._initial = None         # the initial state of the FTS
    self.HIDDEN_DEADLOCKS = None # the list of hidden deadlocks in the FTS
    self.DEAD_TRANSITIONS = None # the list of dead transitions in the FTS


  def state(self, id):
    """this method adds the state $id to the FTS"""
    res = self._nodes.get(id)
    if(res is None):
      data = self.node_default_data(id)
      res = State(id, data)
      self._nodes[id] = res
    return res

  def transition(self, _in, _out, label, constraint):
    """this method adds the transition between the states $_in and $_out, with the label $label and the constraint $constraint to the FTS"""
    _in = self.state(_in)
    _out = self.state(_out)
    edge = Transition(container(_c=constraint, _label=label, DEAD=True, FALSE_OPT=(constraint != True)), _in, _out)
    _in._connections[1].add(edge)
    _out._connections[0].add(edge)
    self._connections.append(edge)

  # virtual
  def node_default_data(self, id): return container(HIDDEN=False, _c=None)

  # this method sets the initial state of the FTS
  def initial_state(self, id): 
    self._initial = self.state(id)
    return self._initial

  def is_live(self): return len(self.HIDDEN_DEADLOCKS) == 0

  # method to compute the status of nodes and transition in the FTS
  def analyse(self):
    if(self._initial == None): raise Exception("the initial state of the FTS ")
    else:
      self.analyse_implem()
      self.HIDDEN_DEADLOCKS = [n for n in self._nodes.values() if n._data.HIDDEN]
      self.DEAD_TRANSITIONS = [c for c in self._connections if c._data.DEAD]
  def analyse_implem(self): pass # abstract method, to be implemented

  def dot(self, name):
    res = f"""digraph G {{ # the FTS is encoded as a directed graph
  # the following three is only used for display
  node [shape=circle];
  rankdir=LR;
  FeatureModel [shape=plaintext, style=filled, color=yellow, label="{True}"];


  FM="{True}";    # the feature model of the FTS
  name="{name}"; # the name of the FTS
  {self._initial._id} [initial=True]

  """
    for e in self._connections:
      res = res + f"\n  {e.get_in()._id} -> {e.get_out()._id} [label=\"{e._data._label} | True\"]"
    return res + '\n}\n'




#
# TEST PATH
#


class FTS_TEST_PATH_TRAVERSAL(FTS):
  def __init__(self, fm):
    super(FTS_TEST_PATH_TRAVERSAL, self).__init__(fm)
  def analyse_implem(self):
    #print(list(map(lambda e: f"({e.get_in()._id}, {e.get_out()._id})", self._connections)))
    for n in self._nodes.values():
      n._visited = False
      n._data._c = 0
    self._step(self._initial)
    print("FINAL STATE")
    for n in self._nodes.values():
      print("{}: {}".format(n._id, n._data._c))

  def _step(self, node):
    if(not node._visited):
      #print(node._id)
      node._data._c = node._data._c + 1
      node._visited = True
      for e in node.nexts():
        #print(f"traversing transition ({e.get_in()._id}, {e.get_out()._id})")
        self._step(e.get_out())
      node._visited = False





#
# BENCH PATH
#


class FTS_JOURNAL2(FTS):
  __slots__ = "_limits"
  def __init__(self, fm):
    super(FTS_JOURNAL2, self).__init__(fm)

  def node_default_data(self, id): return container(HIDDEN=False, _c=0, waiting=0, roots=None, _local_c=0)

  def analyse_implem(self):
    # initialize the node's data
    #self._node_data = {}

    #self._initial._data._local_c = 1
    todoG = { ((self._initial,), 1) }
    while(todoG):
      roots, subg_init = todoG.pop() # get a subgraph specification
      self._tag_subg(roots, subg_init) # annotate the subgraph
      ## TEST: computing the paths
      #print("GRAPH for roots: {}".format(", ".join(n._id for n in roots)))
      #for n in self._nodes.values():
      #  print("  {}: {}".format(n._id, n._data._c.waiting))
      ## END TEST
      initial = roots[-1]
      #self._doneN = set(roots)
      next_nodes, waiting_nodes = {initial}, set() #self._topological_next(initial)
      while(next_nodes):
        node = next_nodes.pop()
        # COMPUTATION
        #self._doneN.add(node)
        new_nexts, new_waiting = self._topological_next(node)
        #print(f"Analysing {node._id} => {[n._id for n in new_nexts]}, {[n._id for n in new_waiting]}")
        # TEST: computing the paths
        #print("managing node {} => {}, {} | {}".format(next_node._id, [n._id for n in new_nexts], [n._id for n in new_waiting], next_node._data._c.waiting_ndata))
        # END TEST
        next_nodes.update(new_nexts)
        waiting_nodes.difference_update(new_nexts)
        waiting_nodes.update(new_waiting)
      for n in waiting_nodes:
        print(f"planning graph for node {n._id} => {n._data._local_c}")
        todoG.add( (roots + (n,), n._data._local_c) )
        n._data._local_c = 0
      #exit(0)
    # TEST: computing the paths
    print("FINAL STATE")
    for n in self._nodes.values():
      n._data._c = n._data._c + n._data._local_c
      print("{}: {}".format(n._id, n._data._c))
    # END TEST

  def _topological_next(self, node):
    new_nodes = set()
    for e in node.nexts():
      n = e.get_out()
      if(n._data.waiting is not self._limits):
        n._data.waiting = n._data.waiting - 1
        new_nodes.add(n)
        n._data._local_c = n._data._local_c + node._data._local_c # computation
    res = [], []
    for n in new_nodes: res[n._data.waiting is not 0].append(n)
    return res


# identify the subgraph, then parse it, adding a dependency for every node. Here, for efficiency, visited = roots, and we compute waiting. then doneN should be changed in the rest, so a node n is not in nexts and waiting at the same time
  def _tag_subg(self, roots, subg_init):
    self._limits = roots
    for n in roots: # identify the limits of the subgraph
      n._data.waiting = roots
    initial = roots[-1]
    #print(f"creating graph for {[n._id for n in roots]}")
    for e in initial.nexts(): self._tag_subg_step(e.get_out())
    initial._data._c = initial._data._c + initial._data._local_c
    initial._data._local_c = subg_init

  def _tag_subg_step(self, node):
    if(node._data.waiting is self._limits): return
    elif(node._visited is self._limits):
      node._data.waiting = node._data.waiting + 1
      #print(" GRAPH VISIT: entering node {} => {}".format(node._id, node._data._c.waiting))
    else:
      node._visited = self._limits
      node._data.waiting = 1
      node._data._c = node._data._c + node._data._local_c
      node._data._local_c = 0
      #print(" GRAPH VISIT: entering node {} => 1".format(node._id))
      for e in node.nexts(): self._tag_subg_step(e.get_out())


#
#
#class FTS_JOURNAL_reachable2(object):
#  __slots__ = "_data",
#  def visit(self, g, roots, subg_init):
#    self._data = roots
#    #print("GENERATING GRAPH for roots: {}".format(", ".join(n._id for n in roots)))
#    initial = roots[-1]
#    initial._data.roots = roots
#    initial._data.waiting = 0
#    #initial._data._c = initial._data._c + initial._data.waiting_ndata - subg_init
#    initial._data._local_c = subg_init
#    for n in g._nodes.values(): n._visited = False
#    for e in initial.nexts(): self._step(e.get_out())
#    for n in g._nodes.values(): n._visited = False
#
#  def _step(self, node):
#    if(node in self._data):
#      #print(" GRAPH VISIT: entering node {} => True".format(node._id))
#      return False
#    elif(node._visited):
#      node._data.waiting = node._data.waiting + 1
#      #print(" GRAPH VISIT: entering node {} => {}".format(node._id, node._data._c.waiting))
#    else:
#      node._visited = True
#      node._data.roots = self._data
#      node._data.waiting = 1
#      node._data._c = node._data._c + node._data._local_c
#      node._data._local_c = 0
#      #print(" GRAPH VISIT: entering node {} => 1".format(node._id))
#      for e in node.nexts(): self._step(e.get_out())
#
#
#




