#!/usr/bin/python

__author__ = "Michael Lienhardt"
__copyright__ = "Copyright 2019, Michael Lienhardt"
__license__ = "GPL3"
__version__ = "1"
__maintainer__ = "Michael Lienhardt"
__email__ = "michael lienhardt@onera.fr"
__status__ = "Prototype"


import gzoumUtils

"""
This file provides a simple implementation of a graph where edges and nodes are generic:
 their class is specified during graph initialization.
Together with this graph implementation is a generic class for graph traversal which implements
 - depth-first search
 - breadth-first search
 - path search
 - topological order
 in directed and undirected graphs (topological order does not make sense in undirected or cyclic graphs)
"""

#
# SIMPLE HYPEREDGE IMPLEMENTATION
#


class HEdge(object):
  """Generic edge class specification"""
  __slots__ = '_nodes', '_data', '_visited'
  def __init__(self, data, *nodes):
    self._nodes = nodes
    self._data = data
    self._visited = False

  def get_data(self): return self._data
  def get_nodes(self): return self._nodes
  # used for directed edges or display
  def get_ins(self): pass
  def get_outs(self): pass
  def get_links(self):
    ins = self.get_ins()
    outs = self.get_outs();
    return [(_in, _out) for _in in ins for _out in outs]

class Edge(HEdge):
  """Simple implementation of the generic edge class"""
  __slots__ = '_ins', '_outs'
  def __init__(self, data, node_in, node_out):
    HEdge.__init__(self, data, node_in, node_out)
    self._ins = [node_in]
    self._outs = [node_out]
  def get_ins(self): return self._ins
  def get_outs(self): return self._outs


class Node(object):
  """Simple implementation of a node class"""
  __slots__ = '_id', '_data', '_visited', '_connections'
  def __init__(self, id, data):
    self._id = id
    self._data = data
    self._visited = False
    self._connections = set()
  def prevs(self): return [ e for e in self._connections if self in e.get_outs()]
  def nexts(self): return [ e for e in self._connections if self in e.get_ins()]


#
# SIMPLE HYPERGRAPH IMPLEMENTATION
#


class Graph(object):
  __slots__ = '_nodes', '_connections', '_node_class', '_edge_class'

  def __init__(self, node_class, edge_class):
    self._node_class = node_class # virtual nodes
    self._edge_class = edge_class # virtual edges
    self._nodes = {}
    self._connections = []


  def node(self, id, data=None):
    """adds a node $id of data $data in the graph"""
    res = self._nodes.get(id)
    if(res == None):
      data = data if(data != None) else self.node_default_data(id)
      res = self._node_class(id, data)
      self._nodes[id] = res
    return res

  def edge(self, data, *args):
    """adds an edge tagged with $data between the nodes with id in $args"""
    nodes = list(map(lambda id: self.node(id), args))
    edge = self._edge_class(data, *nodes)
    for node in nodes:
      node._connections.add(edge)
    self._connections.append(edge)

  # virtual
  def node_default_data(self, id): return None


class GraphVisitor(object):
  kind = gzoumUtils.container(dfs=0, path=1, bfs=2)

  #def visit(self, graph, kind=GraphVisitor.kind.dfs, directed=True, root=None):
  def visit(self, graph, kind=0, directed=True, roots=None, wait_deps=False):
    if(directed):
      def next_node(node): return node.nexts()
      def next_edge(edge): return edge.get_outs()
    else:
      def next_node(node): return node._connections
      def next_edge(edge): return edge._nodes

    if(wait_deps):
      tmp_node = next_node
      def next_node(node): return [ e for e in tmp_node(node) if(all(n._visited for n in e.get_ins())) ]
      tmp_edge = next_edge
      def next_edge(edge): return [ n for n in tmp_edge(edge) if(all(e._visited for e in n.prevs())) ]

    if(kind == GraphVisitor.kind.dfs):
      def step(node):
        node._visited = True
        if(self.enter_node(node)):
          for e in next_node(node):
            if(not e._visited):
              e._visited = True
              if(self.enter_edge(e)):
                for node2 in next_edge(e):
                  if(not node2._visited): step(node2)
                self.exit_edge(e)
          self.exit_node(node)
    elif(kind == GraphVisitor.kind.path):
      def step(node):
        node._visited = True
        if(self.enter_node(node)):
          for e in next_node(node):
            if(not e._visited):
              e._visited = True
              if(self.enter_edge(e)):
                for node2 in next_edge(e):
                  if(not node2._visited): step(node2)
                self.exit_edge(e)
              e._visited = False
          self.exit_node(node)
        node._visited = False    
    elif(kind == GraphVisitor.kind.bfs):
      def step(node):
        to_see = [node]
        seen = []
        while(len(to_see) > 0):
          node = to_see.pop(0)
          edges = []
          node._visited = True
          if(self.enter_node(node)):
            for e in next_node(node):
              if(not e._visited):
                e._visited = True
                edges.append(e)
                if(self.enter_edge(e)):
                  for node2 in next_edge(e):
                    if(not node2._visited): to_see.append(node2)
            seen.append((node, edges))
        while(len(seen) > 0):
          node, edges = seen.pop(-1)
          for e in edges: self.exit_edge(e)
          self.exit_node(node)

    if(roots != None):
      for root in roots: step(graph._nodes[root])
    else:
      for n in graph._nodes.values():
        if(not n._visited): step(n)
    
    if(kind != GraphVisitor.kind.path):
      for n in graph._nodes.values(): n._visited = False
      for e in graph._connections: e._visited = False

  # virtual methods
  def enter_node(self, node): return True
  def enter_edge(self, edge): return True
  def exit_edge(self, edge): return True
  def exit_node(self, node): return True
