#!/usr/bin/python

__author__ = "Michael Lienhardt"
__copyright__ = "Copyright 2019, Michael Lienhardt"
__license__ = "GPL3"
__version__ = "1"
__maintainer__ = "Michael Lienhardt"
__email__ = "michael lienhardt@onera.fr"
__status__ = "Prototype"


from gzoumUtils import element, visitor, container, add_method

#
# SIMPLE PREDICATE CONSTRAINT IMPLEMENTATION
# SPC stands for "Simple Predicate Constraint"
#


class SPC(element):
  __slots__ = "_content"
  def __init__(self, *args): self._content = args

@add_method(SPC, 'check')
def SPC_check(self, model): return SPC_truthVisitor(model).visit(self)
@add_method(SPC, 'toString')
def SPC_toString(self, debug=False):
  if(debug): return SPC_toStringDebugVisitor().visit(self)
  else: return SPC_toStringVisitor().visit(self)
@add_method(SPC, 'getVariables')
def SPC_getVariables(self): return SPC_getVariablesVisitor().visit(self)
@add_method(SPC, 'subst')
def SPC_subst(self, subst): return SPC_SubstitutionVisitor(subst).visit(self)

class And(SPC): pass
class Xor(SPC): pass
class Or(SPC): pass
class Not(SPC):
  def __init__(self, arg): self._content = arg
class Implies(SPC):
  def __init__(self, left, right): SPC.__init__(self, left, right)
class Iff(SPC):
  def __init__(self, left, right): SPC.__init__(self, left, right)

class SPC_toStringDebugVisitor(visitor):
  def visit(self, el): return self.manage_sub(el)
  def manage_sub(self, sub): return sub.accept(self) if isinstance(sub, SPC) else str(sub)

  def And(self, el): return "And({})".format(", ".join([ self.manage_sub(sub) for sub in el._content ]))
  def Xor(self, el):  return "Xor({})".format(", ".join([ self.manage_sub(sub) for sub in el._content ]))
  def Or(self, el):  return "Or({})".format(", ".join([ self.manage_sub(sub) for sub in el._content ]))
  def Not(self, el): return "Not({})".format(self.manage_sub(el._content))
  def Implies(self, el): return "Implies({}, {})".format(self.manage_sub(el._content[0]), self.manage_sub(el._content[1]))
  def Iff(self, el): return "Iff({}, {})".format(self.manage_sub(el._content[0]), self.manage_sub(el._content[1]))

#
# ADD Z3 SUPPORT
#

try:
  import z3
  z3_loaded = True
except ImportError:
  z3_loaded = False

if(z3_loaded):

  class SPC_toZ3Visitor(visitor):
    """This visitor stranslates an SPC constraint into a Z3 constraint"""
    __slots__ = "_var_map"
    def __init__(self):
      self._var_map = {}
  
    def manage_sub(self, sub):
      #print(str(sub))
      if(isinstance(sub, str)):
        res = self._var_map.get(sub)
        if(res == None):
          res = z3.Bool(sub)
          self._var_map[sub] = res
        return res
      elif(isinstance(sub, bool)): return sub
      else: return sub.accept(self)
  
    def And(self, el): return z3.And(tuple( self.manage_sub(sub) for sub in el._content ))
    def Or(self, el): return z3.Or(tuple( self.manage_sub(sub) for sub in el._content ))
    def Xor(self, el):
      subs = tuple( self.manage_sub(sub) for sub in el._content )
      if(len(subs) is 0): return False
      elif(len(subs) is 1): return subs[1]
      else:
        res = [z3.Or(subs)]
        for i in range(len(subs)):
          res.append(z3.Implies(subs[i], z3.Not(z3.Or(tuple( sub for j,sub in enumerate(subs) if(j is not i) )))))
        return z3.And(res)
    def Not(self, el): return z3.Not(self.manage_sub(el._content))
    def Implies(self, el): return z3.Implies(self.manage_sub(el._content[0]), self.manage_sub(el._content[1]))
    def Iff(self, el): return self.manage_sub(el._content[0]) == self.manage_sub(el._content[1])

  """The following method, added to the SPC class, gives if the constraint is satisfiable or not"""
  @add_method(SPC)
  def hasZ3Models(c):
    z3_constraint = SPC_toZ3Visitor().visit(c)
    solver = z3.Solver()
    solver.add(z3_constraint)
    return (solver.check() == z3.sat)

  

